"""
{base}, export layers from CadSoft Eagle files
Copyright (C) 2013  Tomaz Solc <tomaz.solc@tablix.org>
Copyright (C) 2015  Bernard Pratz <guyzmo+pea@m0g.net>

Usage: {base} {command} <input> <type> [<output>:<layer> ...]

Options:
  <input>               .brd, .sch or .lbr file to extract data from
  <type>                chosen output type
  <output>              filename to export data to
  <layer>               loyer to export data from, linked with the output file

<type> can be any of:  {types}
<layer> can be any of: {layers}

"""

from eagle_automation.config import config
import os
import re
import sys
import docopt
import tempfile
import itertools
import subprocess

def ranges(i):
    for a, b in itertools.groupby(enumerate(i), lambda t: t[1] - t[0]):
        b = list(b)
        yield b[0][1], b[-1][1]


def get_extension(path):
	return path.split(os.extsep)[-1].lower()

class BadExtension: Exception


out_types=dict()

class EagleScriptExport:
	def __init__(self, workdir=None, verbose=False):
		self.workdir = workdir
		self.verbose = verbose

	def export(self, in_path, layers, out_paths):

		open(in_path, "rb").close()

		extension = get_extension(in_path)
		script = self.write_script(extension, layers, out_paths)

		for out_path in out_paths:
			# to stop Eagle trowing up dialogs that
			# files already exist
			try:
				os.unlink(out_path)
			except OSError:
				pass

		script += ['QUIT']

		script_string = ';'.join(script)

		cmd = [config.EAGLE, "-C" + script_string, in_path]
		subprocess.call(cmd)

		self.clean()

	def clean(self):
		pass

class EaglePNGExport(EagleScriptExport):
	_page = 1

	def set_page(self, page):
		self._page = page

	def write_script(self, extension, layers, out_paths):

		script = []

		if extension == 'brd':
			script += [	"DISPLAY ALL",
					"RATSNEST"
				]
		elif extension == 'sch':
			script += [	"EDIT .s%d" % self._page ]
		else:
			raise BadExtension

		for layer, out_path in zip(layers, out_paths):
			assert out_path.endswith(".png")

			script += [	"DISPLAY None",
					"DISPLAY %s" % (' '.join(layer['layers']),),
					"EXPORT IMAGE %s MONOCHROME %d" % (out_path, config.DPI)
				]

		return script


out_types['png'] = EaglePNGExport

class EagleBOMExport(EagleScriptExport):
	ULP_TEMPLATE_HEAD = ""
	ULP_TEMPLATE_TAIL = ""
	ULP_TEMPLATE = r"""
		schematic(SCH) {

			string FileName;
			string json;
			string sep = "";

			FileName = filesetext("%(out_path)s", ".json");

			output(FileName, "wt") {
				printf("{\n");
				printf("\t\"items\": [\n");
				SCH.parts(P) {
					if (P.device.package) {
						json = sep + "\t\t{"
							+   "\"prefix\": \""      + P.device.prefix   + "\", "
							+   "\"designator\": \""  + P.name            + "\", "
							+   "\"value\": \""       + P.value           + "\", "
							+   "\"description\": \"" + P.device.headline + "\", "
							+   "\"package\": \""     + P.device.package.name  + "\" "
							+ "}";
						sep = ",\n";
						printf("%%s", json);
					}
				}
				printf("\n\t]\n}\n");
			}
		}
		"""

	def collapse_bom(self):
		import json
		import csv

		# out_bom[prefix][package][value] -> [devices]
		out_bom = dict()
		with open(os.path.join(self.ulp_dir, "bom.json"), 'r') as bom:
			bom = json.load(bom)
			for bom_path in self.bom_path:
				for part in bom['items']:
					prefix, package, value = part['prefix'], part['package'], part['value']
					out_bom.setdefault(prefix, dict()).setdefault(package, dict()).setdefault(value, list()).append(part)

				if '.json' in bom_path:
					with open(bom_path, 'w') as out:
						out.write(json.dumps(out_bom))

				elif '.csv' in bom_path:
					with open(bom_path, 'w', newline='') as csvfile:
						bom_writer = csv.writer(csvfile, dialect='excel', delimiter='	', quotechar='"', quoting=csv.QUOTE_MINIMAL)
						bom_writer.writerow(['Prefix', 'Packaging', 'Value', 'Nb', 'Devices', 'Description'])
						# d['R']['SMT0805']['0R'][1]
						for prefix, packages in out_bom.items():
							for package, items in packages.items():
								for value, devices in items.items():
									range_list = ranges([int(re.sub(r'[a-zA-Z]*', r'', d['designator'])) for d in devices])
									row = [prefix,
											package,
											value,
											len(devices),
											",".join(["{}-{}".format(x,y) if x != y else str(x) for x,y in range_list]),
											devices[0]['description']]
									bom_writer.writerow(row)

				elif '.xlsx' in bom_path:
					try:
						from xlsxwriter.workbook import Workbook
					except ImportError:
						print("Please install xlsxwriter: `pip install xlsxwriter`")
						sys.exit(2)
					try:
						workbook = Workbook(bom_path)
						title = workbook.add_format({'bold': True, 'bg_color': 'gray'})
						bom_writer = workbook.add_worksheet()
						bom_writer.set_column(0, 0, 4.50)
						bom_writer.set_column(1, 1, 14.50)
						bom_writer.set_column(2, 2, 16)
						bom_writer.set_column(3, 3, 2)
						bom_writer.set_column(4, 5, 50.00)
						keys = ['Prefix', 'Packaging', 'Value', 'Nb', 'Devices', 'Description']
						for c, key in enumerate(keys):
							bom_writer.write_string(0, c, key, title)
						row_idx =1
						for prefix, packages in out_bom.items():
							for package, items in packages.items():
								for value, devices in items.items():
									range_list = ranges([int(re.sub(r'[a-zA-Z]*', r'', d['designator'])) for d in devices])
									range_list = ",".join(["{}-{}".format(x,y) if x != y else str(x) for x,y in range_list])
									bom_writer.write(row_idx, 0, prefix)
									bom_writer.write(row_idx, 1, package)
									bom_writer.write(row_idx, 2, value)
									bom_writer.write(row_idx, 3, str(len(devices)))
									bom_writer.write(row_idx, 4, range_list)
									bom_writer.write(row_idx, 5, devices[0]['description'])
									row_idx += 1
					finally:
						workbook.close()
				elif '.xls' in bom_path:
					print("TODO xls support!")

	def write_script(self, extension, layers, out_paths):
		if extension != 'sch':
			raise BadExtension

		self.ulp_dir = self.workdir or tempfile.mkdtemp()
		self.ulp_path = os.path.join(self.ulp_dir, "bom.ulp")

		ulp = open(self.ulp_path, "w")
		ulp.write(self.ULP_TEMPLATE_HEAD)

		self.bom_path = []
		for layer, out_path in zip(layers, out_paths):
			self.bom_path.append(out_path)

		assert '"' not in out_path
		ulp.write(self.ULP_TEMPLATE % {
			'out_path': os.path.join(self.ulp_dir, "bom.json"),
		})

		ulp.write(self.ULP_TEMPLATE_TAIL)
		ulp.close()

		if self.verbose:
			print("Info: Script path:", self.ulp_path)
			print("Info: Raw bom output in:", os.path.join(self.ulp_dir, "bom.json"))

		return [
            "DISPLAY ALL",
            "RUN %s" % (self.ulp_path,)
        ]

	def clean(self):
		self.collapse_bom()
		os.unlink(os.path.join(self.ulp_dir, "bom.json"))
		os.unlink(self.ulp_path)
		if not self.workdir:
			os.rmdir(self.ulp_dir)


out_types['bom'] = EagleBOMExport


class EagleDirectoryExport(EagleScriptExport):

	def write_script(self, extension, layers, out_paths):

		if extension != 'lbr':
			raise BadExtension

		script = [	"EXPORT DIRECTORY %s" % out_paths[0] ]

		return script

class EaglePDFExport(EagleScriptExport):
	def write_script(self, extension, layers, out_paths):

		script = []

		if extension == 'brd':
			script += [	"DISPLAY ALL",
					"RATSNEST"
				]
		else:
			raise BadExtension

		for layer, out_path in zip(layers, out_paths):

			ll = set(layer['layers']) | set(config.DOCUMENT_LAYERS)

			script += [	"DISPLAY None",
					"DISPLAY %s" % (' '.join(ll),),
					"PRINT FILE %s BLACK SOLID" % (out_path,),
				]

		return script


out_types['pdf'] = EaglePDFExport


class EagleMountSMDExport(EagleScriptExport):

	# Following ULP code based on "mountsmd.ulp" by CadSoft

	ULP_TEMPLATE_HEAD = """
		board(B) {
		string fileName;
	"""

	ULP_TEMPLATE = """
		fileName = "%(out_path)s";
		output(fileName) {

			B.elements(E) {

				int wasSmd,
					xmax =-2147483648,
					xmin = 2147483647,
					ymax = xmax,
					ymin = xmin;

				wasSmd = 0;

				E.package.contacts(C) {
					if (C.smd && C.smd.layer == %(pp_id)d) {
						wasSmd = 1;

						if (C.x > xmax) xmax = C.x;
						if (C.y > ymax) ymax = C.y;
						if (C.x < xmin) xmin = C.x;
						if (C.y < ymin) ymin = C.y;
					}
				}

				if (wasSmd)
					printf("%%s %%5.2f %%5.2f %%3.0f %%s %%s\\n",
					E.name, u2mm((xmin + xmax)/2), u2mm((ymin + ymax)/2),
					E.angle, E.value, E.package.name);
			}
		}
	"""

	ULP_TEMPLATE_TAIL = """
	}
	"""

	def write_script(self, extension, layers, out_paths):

		if extension != 'brd':
			raise BadExtension

		self.ulp_dir = self.workdir or tempfile.mkdtemp()
		self.ulp_path = os.path.join(self.ulp_dir, "export.ulp")

		ulp = open(self.ulp_path, "w")
		ulp.write(self.ULP_TEMPLATE_HEAD)

		for layer, out_path in zip(layers, out_paths):

			pp_id = layer['pp_id']

			assert '"' not in out_path
			ulp.write(self.ULP_TEMPLATE % {
				'out_path': out_path,
				'pp_id': pp_id
			})

		ulp.write(self.ULP_TEMPLATE_TAIL)
		ulp.close()

		if self.verbose:
			print(self.ulp_path)

		return [	"DISPLAY ALL",
				"RUN %s" % (self.ulp_path,) ]

	def clean(self):
		os.unlink(self.ulp_path)
		if not self.workdir:
			os.rmdir(self.ulp_dir)


out_types['mountsmd'] = EagleMountSMDExport


class EagleCAMExport:
	def __init__(self, workdir=None, verbose=False):
		self.verbose = verbose

	def export(self, in_path, layers, out_paths):

		open(in_path, "rb").close()

		extension = get_extension(in_path)
		if extension != 'brd':
			raise BadExtension

		for layer, out_path in zip(layers, out_paths):
			options = ["-X", "-d" + self.DEVICE, "-o"  + out_path]
			if layer.get('mirror'):
				options.append("-m")
			cmd = [config.EAGLE] + options + [in_path] + layer['layers']
			subprocess.call(cmd)

class EagleGerberExport(EagleCAMExport):
	DEVICE = "GERBER_RS274X"


out_types['gerber'] = EagleGerberExport

class EagleExcellonExport(EagleCAMExport):
	DEVICE = "EXCELLON"


out_types['excellon'] = EagleExcellonExport


################################################################################

def export_main(verbose=False):
	args = docopt.docopt(__doc__.format(
		base=sys.argv[0],
		command=sys.argv[1],
		types=', '.join(out_types.keys()),
		layers=', '.join(config.LAYERS.keys())
	))

	if verbose:
		print("Arguments:", args)

	layers = []
	out_paths = []
	for arg in args['<output>:<layer>']:
		try:
			out_path, layer_name = arg.split(":")
		except ValueError:
			out_path = arg
			layer_name = None

		extension = args['<input>'].split('.')[-1].lower()
		if extension == 'brd':
			if layer_name is None:
				print("Layer name required when exporting brd files")
				sys.exit(1)

			try:
				layer = config.LAYERS[layer_name]
			except KeyError:
				print("Unknown layer: " + layer_name)
				sys.exit(1)
		elif extension == 'sch':
			layer = {'layers': ['ALL']}
		else:
			print("Bad extension %s: Eagle requires file names ending in sch or brd" % extension)
			sys.exit(1)

		layers.append(layer)
		out_paths.append(out_path)

	try:
		export_class = out_types[args['<type>']]
	except KeyError:
		print("Unknown type: " + out_type)
		sys.exit(1)

	export_class(verbose=verbose).export(args['<input>'], layers, out_paths)


if __name__ == "__main__":
    export_main()



# Default configuration file for eagle_automation.

# LAYERS dictionary provides mapping between Eagle layers and export layers (as
# used by "eagleexport").
#
# Each export layer consists of one or more Eagle layers and usually
# corresponds to one mask (copper, silkscreen, paste, etc.)
#
# Layer properties:
#
# 'layers' : List of Eagle layers to include on this export layer.
#
# 'pp_id'  : Numerical Eagle layer ID for components placed on this layer. Used
#            when exporting pick&place data.
#
# 'mirror' : Whether to mirror this layer on export.
LAYERS = {
	'topassembly': {
		'layers': ['tPlace', 'tNames', 'tDocu'],
		'pp_id': 1,
	},

	'topsilk': {
		'layers': ['tPlace', 'tNames'],
	},

	'toppaste': {
		'layers': ['tCream'],
	},

	'topmask': {
		'layers': ['tStop'],
	},

	'topcopper': {
		'layers': ['Top', 'Pads', 'Vias'],
	},

	'bottomcopper': {
		'layers': ['Bottom', 'Pads', 'Vias'],
		'mirror': True,
	},

	'bottommask': {
		'layers': ['bStop'],
		'mirror': True,
	},

	'bottompaste': {
		'layers': ['bCream'],
		'mirror': True,
	},

	'bottomsilk': {
		'layers': ['bPlace', 'bNames'],
		'mirror': True,
	},

	'bottomassembly': {
		'layers': ['bPlace', 'bNames', 'bDocu'],
		'mirror': True,
		'pp_id': 16,
	},

	'outline': {
		'layers': ['Milling'],
	},

	'measures': {
		'layers': ['LayerStackup', 'DrillLegend', 'Measures'],
	},

	'drills': {
		'layers': ['Drills', 'Holes'],
	},
}

# Eagle layer names to always include when exporting documentation formats (e.g. PDF)
#
# These layers are usually used to add dimensions, frames, signatures, etc. to
# all exported pages.
DOCUMENT_LAYERS = ['Dimension', 'Document']

# Path to Eagle binary to use.
EAGLE = "eagle"

# DPI for bitmap exports.
DPI = 400



#!/usr/bin/env python

"""pea diff: compare CadSoft Eagle files
Copyright (C) 2015  Bernard Pratz <guyzmo+github@m0g.net>
Copyright (C) 2014  Tomaz Solc <tomaz.solc@tablix.org>

USAGE: {prog} {command} [options] <from-file> <to-file>

Parameters:
    <from-file>     File to diff from
    <to-file>       File to diff to

Options:
    -p,--page       Page to compare on multi-page schematics [default: 0]

"""

from __future__ import print_function

import os
import sys
import docopt
import difflib
import tempfile

from PIL import Image, ImageOps, ImageChops

from eagle_automation.config import config
from eagle_automation.export import get_extension, BadExtension, EaglePNGExport, EagleDirectoryExport

def to_png(in_path, page):

	workdir = tempfile.mkdtemp()

	extension = in_path.split('.')[-1].lower()
	if extension == 'brd':
		layers = config.LAYERS.values()
		out_paths = [	os.path.join(workdir, layer + '.png')
				for layer in config.LAYERS.iterkeys() ]
	elif extension == 'sch':
		layers = [{'layers': ['ALL']}]
		out_paths = [os.path.join(workdir, 'all.png')]
	else:
		os.rmdir(workdir)
		raise BadExtension

	export = EaglePNGExport(workdir=workdir)
	export.set_page(page)
	export.export(in_path, layers, out_paths)

	oim = None
	for i, out_path in enumerate(out_paths):
		im = Image.open(out_path).convert("L")
		if oim is None:
			oim = im
		else:
			oim = Image.blend(oim, im, 1.0/(1.0+i))

		os.unlink(out_path)

	os.rmdir(workdir)

	return oim

def to_txt(in_path):
	workdir = tempfile.mkdtemp()

	out_path = os.path.join(workdir, "out.txt")

	export = EagleDirectoryExport()
	export.export(in_path, None, [ out_path ])

	directory = open(out_path).read()

	os.unlink(out_path)
	os.rmdir(workdir)

	return directory

def diff_visual(from_file, to_file, page):

	a_im = to_png(from_file, page=page)
	b_im = to_png(to_file, page=page)

	# make the sizes equal
	# if a sheet contains the filename, it is updated with the temporary name
	# and may thus change the size of the image
	width  = max( (a_im.size[0], b_im.size[0]) )
	height = max( (a_im.size[1], b_im.size[1]) )
	a_im2 = Image.new( "L", (width,height) )
	a_im2.paste( a_im, (0,0) )
	a_im = a_im2
	a_im2 = None
	b_im2 = Image.new( "L", (width,height) )
	b_im2.paste( b_im, (0,0) )
	b_im = b_im2
	b_im2 = None

	added = ImageOps.autocontrast(ImageChops.subtract(b_im, a_im), 0)
	deled = ImageOps.autocontrast(ImageChops.subtract(a_im, b_im), 0)
	same = Image.blend(a_im, b_im, 0.5)

	deled = ImageOps.colorize(deled, "#000", "#f00")
	added = ImageOps.colorize(added, "#000", "#00f")
	same = ImageOps.colorize(same, "#000", "#777")

	im = ImageChops.add(ImageChops.add(added, deled), same)

	im.show()

def diff_text(from_file, to_file):
	a_txt = to_txt(from_file)
	b_txt = to_txt(to_file)

	a_lines = a_txt.split('\n')
	b_lines = b_txt.split('\n')

	diff = difflib.unified_diff(a_lines, b_lines, fromfile=from_file, tofile=to_file, lineterm='')
	print('\n'.join(list(diff)))

def diff(from_file, to_file, page):
	extension = get_extension(from_file)

	if get_extension(to_file) != extension:
		print("%s: both files should have the same extension" % (to_file,))
		return

	if extension == 'brd':
		diff_visual(from_file, to_file, page)
	elif extension == 'sch':
		diff_visual(from_file, to_file, page)
	elif extension == 'lbr':
		diff_text(from_file, to_file)
	else:
		print("%s: skipping, not a board or schematic" % (from_file,))
		return


################################################################################

def diff_main(verbose=False):
	args = docopt.docopt(__doc__.format(prog=sys.argv[0], command=sys.argv[1]))
    if verbose:
        print("Arguments:", args)

	diff(args['<from-file>'], args['<to-file>'], args['--page'])

if __name__ == '__main__':
	diff_main()



#!/usr/bin/env python

"""\
Python Eagle Automation v{version}
Usage: {base} [--verbose] [-c <name>=<value>] [--help] <command> [<args>...]

Options:
    -C <config>        Give path to configuration file
    -c <name=value>    Change configuration option
    -v,--verbose       Give verbose output
    -h,--help          This message

The commands are the following:
    export  Export data out of the brd/sch/lbr for manufacturing
    drill   Export drill data for manufacturing
    diff    Do a diff

See '{base} help <command>' for more information on a specific command.

Copyright (C) 2015  Bernard Pratz <guyzmo+github@m0g.net>
Copyright (C) 2014  Tomaz Solc <tomaz.solc@tablix.org>
Distributed under GPL license.
"""

from __future__ import print_function
import pkg_resources  # part of setuptools

import sys
import docopt

__version__ = pkg_resources.require("eagle_automation")[0].version

def main():
    args = docopt.docopt(__doc__.format(base=sys.argv[0], version=__version__),
                  version = __version__,
                  options_first=True)

    sys.argv = [sys.argv[0]] + [args['<command>']] + args['<args>']
    if args['<command>'] == 'export':
        import eagle_automation.export
        return eagle_automation.export.export_main(verbose=args['--verbose'])
    elif args['<command>'] == 'drill':
        import eagle_automation.drill
        return eagle_automation.drill.drill_main(verbose=args['--verbose'])
    elif args['<command>'] == 'diff':
        import eagle_automation.diff
        return eagle_automation.diff.diff_main(verbose=args['--verbose'])
    elif args['<command>'] in ['help', None]:
        exit(call(['python', sys.argv[0], '--help']))
    else:
        exit("{} is not a {} command. See 'git help'.".format(args['<command>'], sys.argv[0]))


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -+- encoding: utf-8

from setuptools import setup

import sys

opts = dict()
if sys.platform == 'win32':
	import py2exe
	opts.update(
		dict(
			zipfile=None,
			options={
				'py2exe':{
					'includes': ['docopt'],
					'bundle_files': 1
				}
			},
			console=["eagle_automation/pea.py"],
		))
	# windows=[
	#     {
	#         'script': 'eagle_automation/artool.py',
	#         # 'icon_resources': [(1, 'moduleicon.ico')]
	#     }
	# ],

# We do not need pandoc when installing through pypi
if sys.argv[0] == 'setup.py':
    opts.update(dict(
        long_description_markdown_filename='README.md',
        setup_requires=['setuptools_markdown'],
    ))
else:
    opts.update(dict(
        long_description=open('README.md', 'r'),
    ))

setup(
	name='eagle_automation',
	version='0.1.5',
	description='Simple scripts supporting open hardware development using CadSoft EAGLE',
	license='GPL',
	author='Tomaz Solc, Bernard Pratz',
	author_email='tomaz.solc@tablix.org, guyzmo+pea@m0g.net',
    url='https://github.com/guyzmo/eagle-automation',
	packages=['eagle_automation'],
	classifiers=[
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 2',
		'Programming Language :: Python :: 2.7',
		'Development Status :: 4 - Beta',
		'License :: OSI Approved',
		'Operating System :: Unix',
	],
	install_requires=[
		'pillow',
		'docopt',
		'setuptools',
	],
	entry_points="""
	# -*- Entry points: -*-
	[console_scripts]
	pea = eagle_automation.pea:main
	""",
    **opts
)

# offer a nice message when doing pip install
if sys.argv[0] == 'setup.py' and 'install' in sys.argv:
    print("🍻  To start using this tool, do `pea --help`")
